--[[

  XBar for WoW (World of Warcraft)

  By Dr. Doom (Shalune on Kargath)  
]]--

--Global variable so that other mod creators can test to see if XBAR is loaded.
XBAR_ENABLED = true;

--Global vars
XBarModData = { };
XBarData = { };
XBarPlayerName = nil;
XBarOptionSet = "optionset-default";

local XBarInitialized=false;
local ver="1.08"; -- Mod version
local dbver="2";  -- Data format version

--==XBAR FUNCTIONS==--

function XBar_OnLoad()
	this:RegisterEvent("PLAYER_ENTERING_WORLD");

	SLASH_XBAR1 = "/xbar";
	SlashCmdList["XBAR"] = function(msg)
		XBar_SlashHandler(string.lower(msg));
	end
end

function XBar_SlashHandler(msg)
	if (strfind(msg, "config")) then
		-- This function cannot be performed in combat.
                if (InCombatLockdown() == nil) then
			if(XBarConfig:IsVisible()) then
				XBarConfig:Hide();
			else
				XBarConfig:Show();
			end
		end
	elseif (strfind(msg, "show")) then
		msg=strsub(msg,6);
		msg=XBar_ValidateModName(msg);
		if (msg ~= "") then
			DEFAULT_CHAT_FRAME:AddMessage(msg..XBAR_MSG1)
			XBar_ShowHide(msg,2);
		end
	elseif (strfind(msg, "hide")) then
		msg=strsub(msg,6);
		msg=XBar_ValidateModName(msg);
		if (msg ~= "") then
			DEFAULT_CHAT_FRAME:AddMessage(msg..XBAR_MSG1)
			XBar_ShowHide(msg,1);
		end
	elseif (strfind(msg, "resetpos")) then
		msg=strsub(msg,10);
		msg=XBar_ValidateModName(msg);
		if (msg ~= "") then
			DEFAULT_CHAT_FRAME:AddMessage(msg..XBAR_MSG6);
			XBar_ResetPos(msg);
		end
	elseif (strfind(msg, "toggle")) then
		local i,e,f;
		local n=XBarData[XBarOptionSet]["nmods"];

		if (XBarData[XBarOptionSet]["toggle"]) then
			XBarData[XBarOptionSet]["toggle"]=false;
		else
			XBarData[XBarOptionSet]["toggle"]=true;
		end

		if (n~=nil) then
			for i=1,n do
				msg=XBarData[XBarOptionSet]["mod"..tostring(i)]
				if (XBarModData[msg]) then
				if ((XBarModData[msg]["loaded"]) and (not InCombatLockdown())) then
					e=getglobal(msg.."ButtonToggleIcon");
					f=getglobal(msg);
					if (XBarData[XBarOptionSet]["toggle"]) then
						if (f:IsVisible()) then
							e:SetTexture("Interface\\AddOns\\XBar\\xbarminus");
						else
							e:SetTexture("Interface\\AddOns\\XBar\\xbarplus");
						end
						XBar_Update(msg);
					else
						e:SetTexture("");
					end
				end
				end
			end
		end
	elseif (strfind(msg, "options")) then
		msg=strsub(msg,9);
		if (msg=="") then
			DEFAULT_CHAT_FRAME:AddMessage("XBar v"..ver);
			DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT8);
			DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT9);
			DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT10);
			DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT11);
			DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT12);
		elseif (strfind(msg,"view")) then
			local i;
			local n=tonumber(XBarData["noptionsets"]);
			
			DEFAULT_CHAT_FRAME:AddMessage("XBar Option sets:");
			for i=1,n do
				msg=tostring(i)..") ";
				if (XBarData[XBarPlayerName]==tostring(i)) then
					msg=msg.."*";
				end
				msg=msg..XBarData["optionset"..tostring(i)];
				DEFAULT_CHAT_FRAME:AddMessage(msg);
			end
		elseif (strfind(msg,"set")) then
			local i;
			local e=false;
			local n=XBarData["noptionsets"];
			
			msg=strsub(msg,5);
			i=tonumber(msg);
			if (i==nil) then
				e=true;
			elseif ((i<1) or (i>n)) then
				e=true;
			else
				-- Valid number, perform action
				XBarData[XBarPlayerName]=tostring(i);
				ReloadUI();
			end
			if (e) then
				DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG7.."'"..msg.."'");
			end
		elseif (strfind(msg,"delete")) then
			local i,t;
			local e=false;
			local n=tonumber(XBarData["noptionsets"]);
			
			msg=strsub(msg,8);
			i=tonumber(msg);
			msg=tostring(i);
			if (i==nil) then
				e=true;
			elseif ((i<2) or (i>n)) then
				e=true;
			elseif (XBarData[XBarPlayerName]==msg) then
				-- Can't delete our current set
				e=true;
			else
				t=XBarData["optionset"..i];
				XBarData["optionset-"..t]=nil;
				XBarData["optionset"..msg]=XBarData["optionset"..tostring(n)];
				XBarData["optionset"..tostring(n)]=nil;
				-- Valid number, perform action
				-- Whether last or not, players who are using it must use default now
				for x,v in pairs(XBarData) do
					if ((x~="dbver") and (strsub(x,1,9)~="optionset") and (type(v)=="string")) then
						-- By process of elimination, this has to be a player's option set
						if (XBarData[x]==msg) then
							XBarData[x]="1"; -- Found one that was set to this option set
						end
					end
				end

				-- If not the last one, overwrite
				if (i<n) then
					-- Deleted and repointed, now clean up.
					for x,v in pairs(XBarData) do
						if ((x~="dbver") and (strsub(x,1,9)~="optionset") and (type(v)=="string")) then
							-- By process of elimination, this has to be a player's option set
							if (XBarData[x]==tostring(n)) then
								XBarData[x]=msg; -- Found one that was set to this option set
							end
						end
					end
				end

				XBarData["noptionsets"]=n-1;
			end
			if (e) then
				DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG7.."'"..msg.."'");
			else
				DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG10.."'"..t.."'");
			end
		elseif (strfind(msg,"new")) then
			msg=strtrim(strsub(msg,5));
			local i;
			local n=tonumber(XBarData["noptionsets"]);
			local f=false;

			-- make sure its not in there first
			for i=1,n do
				if (XBarData["optionset"..tostring(i)]==msg) then
					DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG7..msg);
					f=true;
				end
			end
			if (msg=="") then
				f=true;
			end
			if (not f) then
				n=n+1;
				XBarData["noptionsets"]=n;
				XBarData["optionset"..tostring(n)]=msg;
				XBarData["optionset-"..msg]={
					["nmods"]=0,
					["mods"]={ },
				};
				DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG8..msg..XBAR_MSG9..tostring(n));
			end
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage("XBar v"..ver);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT1);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT2);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT3);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT4);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT5);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT13);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT6);
		DEFAULT_CHAT_FRAME:AddMessage(XBAR_HELPTEXT7);
	end
end

function XBar_OnEvent()
	if (event == "PLAYER_ENTERING_WORLD") then
		-- Find the player name and save it
		XBar_BuildPlayerName();
		XBar_Start();
	end
end

function XBar_BuildPlayerName()
	-- We must setup the player's name and server to save data.
	if (UnitName("player") ~= UNKNOWNBEING and UnitName("player") ~= UNKNOWNOBJECT and UnitName("player") and not XBarPlayerName) then
		XBarPlayerName = UnitName("player").." of "..GetCVar("realmName");
	end
end

function XBar_GetOptionSet()
	return strsub(XBarOptionSet,11);
end

function XBar_Start()
	local resetdb=false;
	local i,n,m,on,os;

	if (XBarInitialized) then
		return; -- Only do this if we haven't done it yet
	end

	if (not XBarData) then
		resetdb=true;
	elseif (not XBarData[XBarOptionSet]) then
		resetdb=true;
	elseif (XBarData["dbver"] ~= dbver) then
		resetdb=true;
	end

	if (resetdb) then
		--Data structure not initialized for this player, so set one up
		XBarData={
			["dbver"]=dbver;
			["optionset-default"] = {};
			["noptionsets"] = 1;
			["optionset1"] = "default";
			[XBarPlayerName] = "1";
		};
		XBarData[XBarOptionSet]={
			["nmods"]=0,     ---# of mods registered, names will be values of mod1, mod2, etc...
			["mods"]={},     ---Collection of mod values
			["toggle"]=true, -- Show the toggles
		};
	end

	-- Find the option set the player wants to use.
	on=XBarData[XBarPlayerName];
	if (on==nil) then
		XBarData[XBarPlayerName]="1";
		os="optionset-default";
	else
		-- Lets validate that option set
		n=XBarData["noptionsets"];
		if ((tonumber(on)>n) or (tonumber(on)<1)) then
			DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG7..tostring(on));
			on=1; -- reset to defaults
		end
		os="optionset-"..XBarData["optionset"..on];
	end

	XBarOptionSet=os;
	DEFAULT_CHAT_FRAME:AddMessage(XBAR_MSG4..XBar_GetOptionSet());

		-- Initialize all saved mods to "not loaded"
	n=XBarData[XBarOptionSet]["nmods"];
	if (n>0) then
		for i=1,n do
			m=XBarData[XBarOptionSet]["mod"..i];
			if (XBarModData[m] ~= nil) then
				XBarModData[m]["loaded"]=false;
			end
		end
	end
	XBarInitialized=true;
end

function XBar_RegisterAddon(mod)
	-- This function is used by plugins to let us know they are present and loaded.
	local n,v1,v2,i;

	-- Validate that the mod has loaded default settings
	if (not XBarModData[mod]) then
		-- Error message
		DEFAULT_CHAT_FRAME:AddMessage(mod..XBAR_MSG5);
	else
		if (XBarModData[mod]["enabled"]) then
			n=XBarData[XBarOptionSet]["nmods"];
			-- Scan to see if it's in there already
			v1=XBar_ValidateModName(mod);
			if (v1=="") then
				--not found, add it to the list
				i=n+1;
				XBarData[XBarOptionSet]["nmods"]=i;
				XBarData[XBarOptionSet]["mod"..i]=mod;
				XBarData[XBarOptionSet]["mods"][mod]={};
			end
			i=false;
			v1=XBarModData[mod]["dbver"];
			if (XBarData[XBarOptionSet]["mods"][mod]==nil) then
				i=true; -- No database at all
				DEFAULT_CHAT_FRAME:AddMessage(mod..XBAR_MSG2);
			elseif (XBarData[XBarOptionSet]["mods"][mod]["dbver"]~=v1) then
				i=true; -- Database version incorrect
				DEFAULT_CHAT_FRAME:AddMessage(mod..XBAR_MSG3);
			end
			-- Create user-space variables to store settings
			if (i) then
				XBarData[XBarOptionSet]["mods"][mod]={
					["dbver"]=v1;
					["x"]=150,
					["y"]=-150,
					["horizontal"]=XBarModData[mod]["dhorizontal"],
					["hidebar"]=XBarModData[mod]["dhidebar"],
					["order"]=XBarModData[mod]["dorder"],
					["scale"]=XBarModData[mod]["dscale"],
					["tooltips"]=XBarModData[mod]["dtooltips"]
				};
				-- Add in space to save the user checkboxes
				n=XBarModData[mod]["nchecks"];
				if (n==nil) then
					n=0;
					XBarModData[mod]["nchecks"]=0;
				else
					for i=1,n do
						v1=XBarModData[mod]["check"..i];
						v2=XBarModData[mod]["dcheck"..i];
						if (v1) then
							XBarData[XBarOptionSet]["mods"][mod][v1]=v2;
						end
					end
				end
				-- Add in space to save the user sliders
				n=XBarModData[mod]["nsliders"];
				if (n==nil) then
					n=0;
					XBarModData[mod]["nsliders"]=0;
				else
					for i=1,n do
						v1=XBarModData[mod]["slider"..i];
						v2=XBarModData[mod]["dslider"..i];
						XBarData[XBarOptionSet]["mods"][mod][v1]=v2;
					end
				end
			end
			--XBar_SetPos(mod); -- Set the mod to the last known coordinates just in case
			XBarModData[mod]["loaded"]=true; -- mark the mod as loaded
			XBar_GenerateButtons(mod);
		else
			XBarModData[mod]["loaded"]=false; -- mark the mod as unloaded
		end
	end
end

function XBar_GenerateButtons(mod)
	local i,n,b,f;

	n=XBarModData[mod]["nbuttons"];
	if (n==0) then
		return;
	end

	b=getglobal(mod);
	f=CreateFrame("Button",mod.."ButtonToggle",nil,"XBarButtonTemplate");
	f:SetPoint("TOPLEFT",b,"TOPLEFT",10,0);
	f:SetWidth("16");
	f:SetHeight("16");
	f:SetAttribute("XBarToggle",mod);
	f:SetScript("OnClick",XBar_Toggle);
	
	if (not InCombatLockdown()) then
		f:Show();
	end

	for i=1,n do
		f=CreateFrame("Button",mod.."Button"..i,b,"XBarButtonTemplate");
		f:SetAttribute("XBarToggle",nil);
		if (i==1) then
			f:SetPoint("LEFT",mod.."ButtonToggle","RIGHT",1,0);
		else
			f:SetPoint("LEFT",mod.."Button"..(i-1),"RIGHT",1,0);
		end
	end
end

function XBar_Update(mod)
	local i,button,z,id,r1,r2;
	local bar=getglobal(mod);

	spelllist=getglobal(mod.."Spells");
	-- Exit this function if the buttons haven't been initialized yet.
	if (getglobal(mod.."Button1")==nil) then
		return;
	end
	if (XBarData[XBarOptionSet]) then
		XBar_Orient(mod);

		-- This function cannot be performed in combat.
		if (InCombatLockdown() == nil) then
			i=XBarModData[mod]["nbuttons"];

			if (i~=nil) then
				for key=1,i do
					button=getglobal(mod.."Button"..key);
					button:Hide();
				end
			end
			bar:Show();
			getglobal(mod.."ButtonToggle"):Show();
			if (XBarData[XBarOptionSet]["mods"][mod]["hidebar"]) then
				bar:Hide();
			end
		end

		i=0;
		-- Look for callback functions
		local f1=getglobal(XBarModData[mod]["ftexint"]);
		local f2=getglobal(XBarModData[mod]["fbuttoncb"]);
		local f3=getglobal(XBarModData[mod]["fbuttonid"]);

		for key,value in pairs(spelllist) do
			if ((value==" ") and (i>0)) then
				-- Spacer
				getglobal(mod.."Button"..i+1):SetAttribute("XBarSpacer",true);
				if (InCombatLockdown() == nil) then
  					getglobal(mod.."Button"..i+1):Show();
				end
			elseif (strsub(value,1,1)=="#") then
				-- Menu header, just ignore
			else
				id = XBar_GetSpellID(value,XBarModData[mod]["booktype"]);

				if (f3~=nil) then
					id=f3(id,value);
				end

				if (id) then
					i = i + 1;
					button = getglobal(mod.."Button"..i);
					if (button:GetAttribute("XBarSpacer")==nil) then
						button:SetAttribute("XBarSpacer",false);
					end

					-- This function cannot be performed in combat.
					if (InCombatLockdown() == nil) then
						button:SetAttribute("button*","type");
						button:SetAttribute("type", "spell");
						button:SetAttribute("spell*", value);
						button:SetAttribute("ctrl-spell*", ATTRIBUTE_NOOP);
					end

	  				texture = GetSpellTexture(id, XBarModData[mod]["booktype"]);
					-- Call texture interceptor if it is there
					if (f1~=nil) then
						texture=f1(texture,value);
					end
					getglobal(mod.."Button"..i.."Icon"):SetTexture(texture);

					button:SetAttribute("XBarID",id);
					button:SetAttribute("XBarVI",key); -- Save the spell list index for addons to see where it corellates to
					button:SetAttribute("XBarBT",XBarModData[mod]["booktype"]);
					-- This function cannot be performed in combat.
					if (InCombatLockdown() == nil) then
						button:Show();
					end
					if (f2~=nil) then
						f2(button,value,i);
					end
					button:SetScale(XBarData[XBarOptionSet]["mods"][mod]["scale"]);
				end
			end
		end
		XBar_UpdateTT(mod);
		XBar_UpdateCooldowns(mod);

		-- This function cannot be performed in combat.
		if ((i == 0) and (InCombatLockdown() == nil)) then
			bar:Hide();
			getglobal(mod.."ButtonToggle"):Hide();
		end
		if (not XBarData[XBarOptionSet]["toggle"]) then
			getglobal(mod.."ButtonToggle"):Hide();
		end
	end
end

function XBar_UpdateTT(mod)
	local n=XBarModData[mod]["nbuttons"];
	local button,i;

	if (n==nil) then
		return;
	end

	for i = 1, n do
		button = getglobal(mod.."Button"..i);
		button:SetAttribute("XBarTT",XBarData[XBarOptionSet]["mods"][mod]["tooltips"]);
	end
end

function XBar_Orient(mod)
	local button;

	if (not XBarModData[mod]["enabled"]) then
		return;
	end

	local n=XBarModData[mod]["nbuttons"];
	local d1, d2, h, o, s, r, sp;

	if (n==nil) then
		return;
	end

	-- This function cannot be performed in combat.
    if ((InCombatLockdown() == nil) and (XBarData[XBarOptionSet])) then
		h=XBarData[XBarOptionSet]["mods"][mod]["horizontal"];
		o=XBarData[XBarOptionSet]["mods"][mod]["order"];
		r= (o == "za");
		if ((not h) and (not r)) then
			d1="TOP";
			d2="BOTTOM";
		elseif ((not h) and r) then
			d1="BOTTOM";
			d2="TOP";
		elseif (h and (not r)) then
			d1="LEFT";
			d2="RIGHT";
		elseif (h and r) then
			d1="RIGHT";
			d2="LEFT";
		else -- Default is horizontal without reversal
			d1="LEFT";
			d2="RIGHT";
			h=true;
			r=false;
		end

		-- Insert spacing
		if (h) then
			h=1;
		else
			h=0;
		end
		if ((h and not r) or (not h and r)) then
			s=1;
		else
			s=-1;
		end
		h=h * s;
		s=(abs(h)-1) * s;

		for i = 1,n do
			button = getglobal(mod.."Button"..i);
			if (button~=nil) then
				button:ClearAllPoints();
				if (button:GetAttribute("XBarSpacer")) then
					sp=10;
				else
					sp=1;
				end
				if (i==1) then
					button:SetPoint(d1, mod.."ButtonToggle", d2, h*sp, s*sp);
				else
					button:SetPoint(d1, mod.."Button"..(i-1), d2, h*sp, s*sp);
				end
			end
		end
	end
end

function XBar_UpdateCooldowns(mod)
	local button,i,bi,bt,v,id,cooldown,start,duration,enable,texture;

	if (not XBarModData[mod]["loaded"]) then
		return;
	end

	local n=XBarModData[mod]["nbuttons"];
	if (n==nil) then
		return;
	end

	local f1=getglobal(XBarModData[mod]["ftexint"]);
	for i = 1, n do
		button = getglobal(mod.."Button"..i);

		bi=getglobal(mod.."Button"..i.."Icon");
		bt=button:GetAttribute("XBarBT");
		if (bt) then
			id=button:GetAttribute("XBarID");
			v=GetSpellName(id,bt);
			if (v) then
  				texture = GetSpellTexture(id, XBarModData[mod]["booktype"]);

				-- Call texture interceptor if it is there
				if (f1~=nil) then
					texture=f1(texture,v);
				end

				bi:SetTexture(texture);
				
				cooldown = getglobal(mod.."Button"..i.."Cooldown");
				start, duration, enable = GetSpellCooldown(id, bt);
				CooldownFrame_SetTimer(cooldown,start, duration, enable);
			end
		end
	end
end

function XBar_UpdateToggle(bar,x,y)
	local fi=getglobal(bar:GetName().."ButtonToggleIcon");
	
	if (fi==nil) then
		return;
	end

	if (XBarData[XBarOptionSet]["toggle"]) then
		if bar:IsVisible() then
			fi:SetTexture("Interface\\AddOns\\XBar\\xbarminus");
		else
			if (not UIParent:IsVisible()) then
				fi:SetTexture("");
			else
				fi:SetTexture("Interface\\AddOns\\XBar\\xbarplus");
			end
		end
	else
		fi:SetTexture("");
	end
end

function XBar_GetSpellID(spell,book)
	local i = 0;
	local curRank=nil;
	local spellName, spellRank;

	repeat
		i = i + 1;
		spellName, spellRank = GetSpellName(i, book);
		if (spellName == spell) then
			curRank = spellRank;
		elseif (curRank) then
			break; -- break if we passed all the spells
		end
	until (spellName==nil);

	if (curRank==nil) then
		i=nil; -- didn't find it
	else
		i=i-1; -- found it, correct i.
	end

	return i;
end

function XBar_SetTooltip(button)
	local id=this:GetAttribute("XBarID");
	local tt=this:GetAttribute("XBarTT");
	local toggle=this:GetAttribute("XBarToggle");

	if (id==nil) then
		id=-1;
		tt=false;
	end

	if ((id >= 0) and tt) then
		GameTooltip_SetDefaultAnchor(GameTooltip, this);
		GameTooltip:SetSpell(id, this:GetAttribute("XBarBT"));
		GameTooltip:Show();
	elseif (toggle ~= nil) then
		GameTooltip_SetDefaultAnchor(GameTooltip, this);
		GameTooltip:SetText(toggle);
		GameTooltip:Show();
	end
end

function XBar_Scale(mod)
	if (not XBarModData[mod]["loaded"]) then
		return;
	end

	local n=XBarModData[mod]["nbuttons"];
	if (n==nil) then
		return;
	end

	for i = 1,n do
		button = getglobal(mod.."Button"..i);
		button:SetScale(XBarData[XBarOptionSet]["mods"][mod]["scale"]);
	end
end

function XBar_ShowHide(mod,state)
	-- state=0: Toggle
	-- state=1: Hide
	-- state=2: Show
	if (not XBarModData[mod]) then
		DEFAULT_CHAT_FRAME:AddMessage(mod);
		return;
	end
	if (not XBarModData[mod]["loaded"]) then
		return;
	end

	-- This function cannot be performed in combat.
	if (InCombatLockdown() == nil) then
		local frame = getglobal(mod);
		if (frame) then
			if(frame:IsVisible()) then
				if (state ~= 2) then
					frame:Hide();
					XBarData[XBarOptionSet]["mods"][mod]["hidebar"] = true;
				end
			else
				if (state ~= 1) then
					frame:Show();
					XBarData[XBarOptionSet]["mods"][mod]["hidebar"] = false;
				end
			end
		end
	end
end

function XBar_ValidateModName(mod)
	local n=XBarData[XBarOptionSet]["nmods"];
	local v=false;
	local s1=strlower(strtrim(mod));
	local s2;

	for i=1,n do
		s2=XBarData[XBarOptionSet]["mod"..i];
		if (strlower(s2)==s1) then
			return s2;
		end
	end
	return "";
end

function XBar_SaveLoc(bar)
	local mod;

	mod=bar:GetName();
	if (mod) then
		if (not XBarModData[mod]["loaded"]) then
			return;
		end

		local p,rt,rp,x,y=bar:GetPoint();
		XBarData[XBarOptionSet]["mods"][mod]["x"] = x;
		XBarData[XBarOptionSet]["mods"][mod]["y"] = y;

		-- More than likely, we are saving due to a move event, so lets move the toggle button to match.
		XBar_UpdateToggle(bar,x,y);
	end
end

function XBar_ResetPos(mod)
	if (not XBarModData[mod]) then
		return;
	end

	local bar=getglobal(mod);
	bar:ClearAllPoints();
	bar:SetPoint("TOPLEFT",nil,"CENTER");
	XBar_SaveLoc(bar);
end

function XBar_SetPos(mod)
	if (not XBarModData[mod]) then
		return;
	end

	local bar=getglobal(mod);
	local x = XBarData[XBarOptionSet]["mods"][mod]["x"];
	local y = XBarData[XBarOptionSet]["mods"][mod]["y"];

	bar:ClearAllPoints();
	bar:SetPoint("TOPLEFT",nil,"TOPLEFT",x,y);
end

function XBar_Toggle()
	local mod=this:GetName();
	local fi=getglobal(mod.."Icon");

	if (fi:GetTexture()~=nil) then
		mod=strsub(mod,-1*strlen(mod),-13); -- cut off the appendix to find the corresponding bar
		XBar_ShowHide(mod,0);
	end
end

--==XBAR CONFIG FUNCTIONS==--

function XBarConfig_OnShow()
	local i,n,mod,b;
	local cfg=this:GetName();

	if (not XBarData) then
		this:Hide();
		return;
	end
	if (not XBarData[XBarOptionSet]) then
		this:Hide();
		return;
	end

	n=XBarData[XBarOptionSet]["nmods"];
	if (n==nil) then
		this:Hide();
		return;
	end

	-- Initialize select fields to nothing
	XBarConfig:SetAttribute("mod","");
	XBarConfig:SetAttribute("checkoption","");
	XBarConfig:SetAttribute("slideroption","");

	-- Start with all fields hidden
	XBarConfigCheckButtonH:Hide();
	XBarConfigCheckButtonO:Hide();
	XBarConfigCheckButtonV:Hide();
	XBarConfigCheckButtonTT:Hide();
	XBarConfigSliderScale:Hide();
	XBarConfigCurCheck:Hide();
	XBarConfigCheckMenu:Hide();
	XBarConfigCurSlider:Hide();
	XBarConfigSliderMenu:Hide();
	XBarConfigSliderMulti:Hide();

	-- Search for loaded mods
	b=false;
	for i=1,n do
		mod=XBarData[XBarOptionSet]["mod"..i];
		if (XBarModData[mod]) then
		if (XBarModData[mod]["loaded"]) then
			b=true;
		end
		end
	end

	--Tell user whether we have anything installed or not.
	if (not b) then
		-- No data, display an error message
		XBarConfigCurMod:SetText(XBAR_NOMOD);
		XBarConfigModMenu:Hide();
	else
		XBarConfigCurMod:SetText(XBAR_SELMOD);
		XBarConfigModMenu:Show();
	end
end

function XBarConfigModMenu_Select()
	local mod=this.value;
	XBarConfigCurMod:SetText(mod);

	if (InCombatLockdown()) then
		return; -- can't do this in combat
	end
	
	XBarConfigCheckButtonH:Show();
	XBarConfigCheckButtonH:SetChecked(XBarData[XBarOptionSet]["mods"][mod]["horizontal"]);

	XBarConfigCheckButtonO:Show();
	XBarConfigCheckButtonO:SetChecked(XBarData[XBarOptionSet]["mods"][mod]["order"]=="za");

	XBarConfigCheckButtonV:Show();
	XBarConfigCheckButtonV:SetChecked(getglobal(mod):IsVisible());

	-- Only show if the author did not disable trackings in the addon
	if (XBarModData[mod]["dtooltips"]) then
		XBarConfigCheckButtonTT:Show();
		XBarConfigCheckButtonTT:SetChecked(XBarData[XBarOptionSet]["mods"][mod]["tooltips"]);
	end

	XBarConfigSliderScale:Show();
	XBarConfigSliderScale:SetValue(XBarData[XBarOptionSet]["mods"][mod]["scale"]);

	-- Show custom checkbox and slider options
	if (XBarModData[mod]["nchecks"]>0) then
		XBarConfigCurCheck:Show();
		XBarConfigCheckMenu:Show();
		XBarConfigCurCheck:SetText(XBAR_SELOPTION);
	else
		XBarConfigCurCheck:Hide();
		XBarConfigCheckMenu:Hide();
		XBarConfigSliderMulti:Hide()
	end

	if (XBarModData[mod]["nsliders"]>0) then
		XBarConfigCurSlider:Show();
		XBarConfigSliderMenu:Show();
		XBarConfigCurSlider:SetText(XBAR_SELOPTION);
	else
		XBarConfigCurSlider:Hide();
		XBarConfigSliderMenu:Hide();
		XBarConfigSliderMulti:Hide()
	end
end

function XBarConfigModMenuFrame_OnLoad()
	-- Loading off the click event from the button itself so we can make sure the data is there.
	-- Reinitializes every time we click the button.
	local info = UIDropDownMenu_CreateInfo();
	local n,i,mod;

	n=XBarData[XBarOptionSet]["nmods"];
	for i=1,n do
		mod=XBarData[XBarOptionSet]["mod"..i]
		if (XBarModData[mod]) then
		if (XBarModData[mod]["loaded"]) then
			info.text = mod;
			info.notCheckable=true;
			info.value = mod;
			info.func = XBarConfigModMenu_Select;
			info.owner = this:GetParent();
			UIDropDownMenu_AddButton(info);
		end
		end
	end
end

function XBarConfigModMenu_OnClick()
	UIDropDownMenu_Initialize(XBarConfigModMenuFrame, XBarConfigModMenuFrame_OnLoad);
	ToggleDropDownMenu(1, nil, XBarConfigModMenuFrame, XBarConfigModMenu, 0, 0);
end

function XBarConfigCheckButtonV_OnClick()
	local mod=XBarConfigCurMod:GetText();
	local bar=getglobal(mod);

	if (this:GetChecked()) then
		XBar_ShowHide(mod,2);
	else
		XBar_ShowHide(mod,1);
	end
end

function XBarConfigCheckButtonH_OnClick()
	local mod=XBarConfigCurMod:GetText();

	if (this:GetChecked()) then
		XBarData[XBarOptionSet]["mods"][mod]["horizontal"] = true;
	else
		XBarData[XBarOptionSet]["mods"][mod]["horizontal"] = false;
	end
	XBar_Orient(mod);
end

function XBarConfigCheckButtonO_OnClick()
	local mod=XBarConfigCurMod:GetText();

	if (this:GetChecked()) then
		XBarData[XBarOptionSet]["mods"][mod]["order"] = "za";
	else
		XBarData[XBarOptionSet]["mods"][mod]["order"] = "az";
	end
	XBar_Orient(mod);
end

function XBarConfigCheckButtonTT_OnClick(mod)
	local mod=XBarConfigCurMod:GetText();

	if (this:GetChecked()) then
		XBarData[XBarOptionSet]["mods"][mod]["tooltips"] = true;
	else
		XBarData[XBarOptionSet]["mods"][mod]["tooltips"] = false;
	end
	XBar_UpdateTT(mod);
end

function XBarConfigSliderScale_OnChange()
	local mod=XBarConfigCurMod:GetText();

	local v=this:GetValue();
	XBarData[XBarOptionSet]["mods"][mod]["scale"] = v;
	XBarConfigSliderScaleVal:SetText(format("%.2f", v));
	XBar_Scale(mod);
end

function XBarConfigCheckMenu_Select()
	local mod=XBarConfigCurMod:GetText();
	local v;
	local c=XBarModData[mod]["check"..this.value];

	if (this.checked) then
		v=false;
	else
		v=true;
	end
	
	XBarData[XBarOptionSet]["mods"][mod][c]=v;
	local f=getglobal(XBarModData[mod]["foptioncb"]);

	if (f~=nil) then
		f(c,v);
	end
end

function XBarConfigCheckMenuFrame_OnLoad(level)
	-- Loading off the click event from the button itself so we can make sure the data is there.
	-- Reinitializes every time we click the button.
	local info = UIDropDownMenu_CreateInfo();
	local n,i,mod;
	local menukey=nil;

	mod=XBarConfigCurMod:GetText();

	n=XBarModData[mod]["nchecks"];
	for i=1,n do
		info.text = XBarModData[mod]["check"..i];
		info.func = XBarConfigCheckMenu_Select;
		info.owner = this:GetParent();
		info.hasArrow=false;
		info.notCheckable=false;
		if (info.text == "") then
			-- Its probably a menu
			info.func=nil;
			info.text = XBarModData[mod]["mcheck"..i];
			info.value = info.text;
			if (info.text ~= nil) then
				menukey=info.text;
				info.hasArrow=true;
				info.notCheckable=true;
				info.checked=false;
				if (level==1) then
					-- Always add the menu if it's the first one.
					UIDropDownMenu_AddButton(info,1);
				end
			end
		elseif (info.text ~= nil) then
			-- Only put something here if we need.
			info.value = i;
			info.checked=XBarData[XBarOptionSet]["mods"][mod][info.text];
			if ((level==2) and (menukey==UIDROPDOWNMENU_MENU_VALUE)) then
				UIDropDownMenu_AddButton(info,2);
			elseif ((level==1) and (menukey==nil)) then
				--haven't encountered any dropdown menus yet
				UIDropDownMenu_AddButton(info,1);
			end
		end
	end
end

function XBarConfigCheckMenu_OnClick()
	UIDropDownMenu_Initialize(XBarConfigCheckMenuFrame, XBarConfigCheckMenuFrame_OnLoad);
	ToggleDropDownMenu(1, nil, XBarConfigCheckMenuFrame, XBarConfigCheckMenu, 0, 0);
end

function XBarConfigSliderMenu_Select()
	local mod=XBarConfigCurMod:GetText();
	local v=this.value;
	local c=XBarModData[mod]["slider"..v];
	local min=XBarModData[mod]["slider"..v.."min"];
	local max=XBarModData[mod]["slider"..v.."max"];
	local fmt=XBarModData[mod]["slider"..v.."format"];
	local v2=XBarData[XBarOptionSet]["mods"][mod][c];

	XBarConfig:SetAttribute("sliderindex",v);
	XBarConfigSliderMultiText:SetText(c);
	XBarConfigSliderMulti:SetMinMaxValues(min,max);
	XBarConfigSliderMulti:SetValueStep(XBarModData[mod]["slider"..v.."step"]);
	XBarConfigSliderMulti:SetValue(v2);
	XBarConfigSliderMultiHigh:SetText(tostring(max));
	XBarConfigSliderMultiLow:SetText(tostring(min));
	XBarConfigSliderMultiVal:SetText(format(fmt, v2));
	XBarConfigSliderMulti:Show();
end

function XBarConfigSliderMenuFrame_OnLoad(level)
	-- Loading off the click event from the button itself so we can make sure the data is there.
	-- Reinitializes every time we click the button.
	local info = UIDropDownMenu_CreateInfo();
	local n,i,mod;
	local menukey=nil;

	mod=XBarConfigCurMod:GetText();

	n=XBarModData[mod]["nsliders"];
	for i=1,n do
		info.text = XBarModData[mod]["slider"..i];
		info.func = XBarConfigSliderMenu_Select;
		info.owner = this:GetParent();
		info.hasArrow=false;
		info.notCheckable=true;
		if (info.text == "") then
			-- Its probably a menu
			info.func=nil;
			info.text = XBarModData[mod]["mslider"..i];
			info.value = info.text;
			if (info.text ~= nil) then
				menukey=info.text;
				info.hasArrow=true;
				if (level==1) then
					-- Always add the menu if it's the first one.
					UIDropDownMenu_AddButton(info,1);
				end
			end
		elseif (info.text ~= nil) then
			-- Only put something here if we need.
			info.value = i;
			if ((level==2) and (menukey==UIDROPDOWNMENU_MENU_VALUE)) then
				UIDropDownMenu_AddButton(info,2);
			elseif ((level==1) and (menukey==nil)) then
				--haven't encountered any dropdown menus yet
				UIDropDownMenu_AddButton(info,1);
			end
		end
	end
end

function XBarConfigSliderMenu_OnClick()
	UIDropDownMenu_Initialize(XBarConfigSliderMenuFrame, XBarConfigSliderMenuFrame_OnLoad);
	ToggleDropDownMenu(1, nil, XBarConfigSliderMenuFrame, XBarConfigSliderMenu, 0, 0);
end

function XBarConfigSliderMulti_OnChanged()
	-- Don't update the value unless it's visible
	if (not this:IsVisible()) then
		return;
	end

	local mod=XBarConfigCurMod:GetText();
	local c=XBarConfigSliderMultiText:GetText();
	local v=this:GetValue();
	local i=XBarConfig:GetAttribute("sliderindex");
	local fmt=XBarModData[mod]["slider"..i.."format"];

	XBarConfigSliderMultiVal:SetText(format(fmt, v));
	XBarData[XBarOptionSet]["mods"][mod][c]=v;
	local f=getglobal(XBarModData[mod]["foptioncb"]);

	if (f~=nil) then
		f(c,v);
	end
end